
# This script is based on jobmanager-lcgpbs provided by LCG. It is modified to
# work with SGE by Javier Fontan (jfontan_AT_gmail.com) from CESGA (www.cesga.es)

# Modify by Pablo Rey (prey@cesga.es) to include a Configuration file

# Modify by Pablo Rey (prey@cesga.es) to add support for DGAS job records. Following Maarten.Litmaath@cern.ch instructions.

# v0.50 by Alvaro Simon (asimon@cesga.es) fixed an issue submitting jobs to different qmaster servers.

# Comments, support and bug reports should be sent to egee-admin_AT_cesga.es

# SGE queues should have an epilog script to delete STDOUT and STDERR files

package Globus::GRAM::JobManager::lcgsge;

@Globus::GRAM::JobManager::lcgsge::ISA = qw(Globus::GRAM::JobManager::Helper);

use Globus::GRAM::Helper;
use Globus::GRAM::Error;
use Globus::GRAM::JobState;
use Globus::GRAM::JobManager;
use Globus::Core::Paths;

use IO::File;
use Config;
use POSIX;

my ($sge_base_path, $sge_bin_path, $sge_qmaster, $mpirun, $qsub, $qstat, $qdel, $qmsg, $cluster, $cpu_per_node, $remote_shell);

BEGIN
{
	# [prey]
	# Including the config file
        my $confile;
        if (defined $ENV{'GLOBUS_LOCATION'}) {
                $confile = $ENV{'GLOBUS_LOCATION'} . "/lib/perl/Globus/GRAM/JobManager/lcgsge.conf";
        } else {
                $confile = "/opt/globus/lib/perl/Globus/GRAM/JobManager/lcgsge.conf";
        }
	do $confile;

	# [jfontan]
	# Change this to sge tools
	#$sge_base_path = '/opt/sge6';
	#$sge_bin_path = $sge_base_path.'/bin/lx24-x86';
	
	# [prey]
	# Getting the $sge_base_path and $sge_bin_path from the config file
	$sge_base_path = $SGE_BASE_PATH;
	# Removing the last / if it is included
	chop($sge_base_path) if ($sge_base_path =~ /\/$/);
	$sge_bin_path = $SGE_BIN_PATH;
	# Removing the last / if it is included
	chop($sge_bin_path) if ($sge_bin_path =~ /\/$/);
        # [ASG]
        # Needed if we have configured a different qmaster server  
        if(defined $SGE_QMASTER && $SGE_QMASTER) {$sge_qmaster = $SGE_QMASTER;}
        else {
        $sge_qmaster = 536;
        }


	$mpirun = $sge_bin_path.'/mpirun';
	$qsub =   $sge_bin_path.'/qsub';
	$qstat =  $sge_bin_path.'/qstat';
	$qdel = $sge_bin_path.'/qdel';
	$qmsg = $sge_bin_path.'/qmsg';
	$cluster = 0;
	$cpu_per_node = 0;
	$remote_shell = '/usr/bin/ssh';
}

sub new
{
	my $proto = shift;
	my $class = ref($proto) || $proto;
	my $self = $class->SUPER::new(@_);

	bless $self, $class;
	return $self;
}

# [prey]
# Added dgas_record function
sub dgas_record
{
	my $jobmanager = 'lcgsge';

	my $queue    = shift;
	my $job_id   = shift;
	my $batch_id = shift;

	my $opt_r = "";

	my $dgas_fd   = $ENV{GATEKEEPER_DGAS_FD};
	my $dgas_path = $ENV{GATEKEEPER_DGAS_PATH};

	my $dgas_add_record = '/opt/lcg/sbin/dgas-add-record';
	my $jm_dgas_helper  = '/opt/lcg/sbin/jm-dgas-helper';

	if (! -x $dgas_add_record) {
		open(LCG, '/etc/sysconfig/lcg');

		while (<LCG>) {
			if (/^LCG_LOCATION=(.+)/) {
				my $lcg = $1;
				$dgas_add_record =~ s-/opt/lcg-$lcg-;
				$jm_dgas_helper  =~ s-/opt/lcg-$lcg-;
				last;
			}
		}

		close LCG;

		if (! -x $dgas_add_record) {
			# out of luck
			return;
		}
	}

	if (!defined($batch_id)) {
		#
		# phase 1
		#
		if (defined($dgas_fd)) {
			system("$jm_dgas_helper $jobmanager $queue >&$dgas_fd");
		}

		if (defined($dgas_path)) {
			$dgas_path =~ s-.*/--;
			$opt_r = "-r" if ($job_id ne "none");
			system("$dgas_add_record $opt_r $dgas_path $job_id");
		}
	} else {
		#
		# phase 2
		#

		system("$dgas_add_record $job_id $batch_id");
	}
}

sub submit
{
	my $self = shift;
	my $description = $self->{JobDescription};
	# [prey]
	# Added	queue variable
	my $queue = $description->queue() || "undefined";
	my ($cpu_time,$wall_time);

	$self->log("Entering sge submit");

	# check jobtype
	if(defined($description->jobtype()))
	{
		#	if($description->jobtype !~ /^(mpi|single|multiple)$/)
		if($description->jobtype !~ /^(single|multiple)$/)
		{
			# [prey]
			# Call to dgas_record function
			dgas_record($queue, "none");
			return Globus::GRAM::Error::JOBTYPE_NOT_SUPPORTED;
		}
	}
	else
	{
		# [prey]
		# Call to dgas_record function
		dgas_record($queue, "none");
		return Globus::GRAM::Error::JOBTYPE_NOT_SUPPORTED;
	}

	if( !defined $description->directory() || $description->directory() eq "")
	{
                # [prey]
                # Call to dgas_record function
                dgas_record($queue, "none");
		return Globus::GRAM::Error::RSL_DIRECTORY();
	}

	if( !defined $description->executable() || $description->executable() eq "")
	{
                # [prey]
                # Call to dgas_record function
                dgas_record($queue, "none");
		return Globus::GRAM::Error::RSL_EXECUTABLE();
	}
	elsif( !defined $description->stdin() || $description->stdin() eq "")
	{
                # [prey]
                # Call to dgas_record function
                dgas_record($queue, "none");
		return Globus::GRAM::Error::RSL_STDIN;
	}

	$self->log("Determining job max time cpu from job description");
	if(defined($description->max_cpu_time())) 
	{
		$cpu_time = $description->max_cpu_time();
		$self->log("   using maxcputime of $cpu_time");
	}
	elsif(! $cluster && defined($description->max_time()))
	{
		$cpu_time = $description->max_time();
		$self->log("   using maxtime of $cpu_time");
	}
	else
	{
		$cpu_time = 0;
		$self->log('   using queue default');
	}

	$self->log("Determining job max wall time limit from job description");
	if(defined($description->max_wall_time()))
	{
		$wall_time = $description->max_wall_time();
		$self->log("   using maxwalltime of $wall_time");
	}
	elsif($cluster && defined($description->max_time()))
	{
		$wall_time = $description->max_time();
		$self->log("   using maxtime of $wall_time");
	}
	else
	{
		$wall_time = 0;
		$self->log('   using queue default');
	}

	foreach ($description->environment())
	{
		if(!ref($_) || scalar(@$_) != 2)
		{
			# [prey]
			# Call to dgas_record function
			dgas_record($queue, "none");
			return Globus::GRAM::Error::RSL_ENVIRONMENT();
		}
	}


	foreach($description->arguments())
	{
		if(ref($_))
		{
                        # [prey]
                        # Call to dgas_record function
                        dgas_record($queue, "none");
			return Globus::GRAM::Error::RSL_ARGUMENTS;
		}
	}

	my $job_id = $self->queue_submit("lcgsge",$cpu_time."|".$wall_time);

	# [prey]
	# Call to dgas_record function
	dgas_record($queue, $job_id) if defined $job_id;

	$self->log("Leaving sge submit");

	return
	{
		JOB_ID => $job_id,
		JOB_STATE => Globus::GRAM::JobState::PENDING
	} if defined $job_id;

	# [prey]
	# Call to dgas_record function
	dgas_record($queue, "none");
	return Globus::GRAM::Error::NO_RESOURCES;
}

sub poll
{
	my $self = shift;
	my $description = $self->{JobDescription};
	my $state;

	my $internal_id = $description->jobid();
	my ($batch_id,$job_submit_time);

	$self->lookup_or_submit(\$batch_id,\$job_submit_time,\$state);
	$self->log("polling job $batch_id") if defined $batch_id;

	my @data;
	my $query_ret = $self->make_a_poll_query(".lcgjm","sgequeue.cache",$job_submit_time,\@data);

	if ( !defined $state )
	{
		my $exit_code = 153;
		my $status_line;

		foreach my $line (@data)
		{
			my $jid;

			if ($line =~ /^(\S+)\s+(\S+)$/)
			{
				($jid,$status_line) = ($1,$2);
				if ($jid eq $batch_id)
				{
					$self->log("  Job found: " . $jid . ' | ' . $status_line );
					$exit_code = 0;
					last;
				}
			}
		}

		# return code 153 = "Unknown Job Id".
		# verifying that the job is no longer there.
		if ($exit_code == 153)
		{
			if ($query_ret)
			{
				$self->log("qstat rc is 153 == Unknown Job ID == DONE");
				$state = Globus::GRAM::JobState::DONE;
			}
			else
			{
				$self->log("Job not found, assuming it is PENDING");
				$state = Globus::GRAM::JobState::PENDING;
			}
		}
		else
		{

			# Get 3rd field (after = )
			$_ = $status_line;

			if(/E|q|w|t/)
			{
				if (/E/)
				{
					$self->cancel();
					$state = Globus::GRAM::JobState::FAILED;
				}
				else
				{
					$state = Globus::GRAM::JobState::PENDING;
				}
			}
			elsif(/s/)
			{
				$state = Globus::GRAM::JobState::SUSPENDED
			}
			elsif(/r/)
			{
				$state = Globus::GRAM::JobState::ACTIVE;
			}
			else
			{
				# This else is reached by an unknown response from pbs.
				# It could be that PBS was temporarily unavailable, but that it
				# can recover and the submitted job is fine.
				# So, we want the JM to ignore this poll and keep the same state
				# as the previous state.  Returning an empty hash below will tell
				# the JM to ignore the respose.
				$self->log("qstat returned an unknown response.  Telling JM to ignore this poll");
				return {};
			}
		}
	}

	#
	# Not using pbsmsg for now.
	#
	#    if ($state == Globus::GRAM::JobState::ACTIVE)
	#    {
	#        $self->l_send_pbsmsg($batch_id);
	#    }

	#
	# Dont append the batch.err, as we're using PBSMSG
	# Not using PBSMSG for now, so allow appending of batch output
	#

	$self->helper_cache_import($internal_id,$batch_id,$job_submit_time,\$state,"batch.out","batch.err");

	return {
		JOB_STATE     => $state
	};
}

sub poll_batch_system
{
	my $self = shift;
	my ($data_ref,$time_ref) = @_;

	my $good_query = 0;
	do
	{
		@$data_ref = ();
		$$time_ref = time();
		local(*JQ);
                # [ASG] We export QMASTER variable if we submit our job to a different server
		if (open(JQ,"export SGE_ROOT=".$sge_base_path." ; export  SGE_QMASTER_PORT=".$sge_qmaster." ; $qstat 2>/dev/null |"))
		{
			my $jid;
			while(<JQ>)
			{
				chomp(my $line = $_);
				# [jfontan]
				# Changed regexp to match SGE qstat output
				if ($line =~ /^ *(\d+) +\d\.\d+ +[^ ]* +[^ ]* +(\w+)/)
				{
					my $st = $2;
					$jid = $1;
                                        # [ASG]
                                        # Check first if $jid is defined
                                        if (defined $jid)
                                        {
					push(@$data_ref,$jid." ".$st);
                                        $jid = undef;
                                        }
				}
			}
			close(JQ);
			$good_query = 1 if $? == 0;
			sleep 30 if !$good_query;
		}
		else
		{
			$good_query = -1;
		}
	} while($good_query != 1);
}

sub cancel_in_batch_system
{
	my $self = shift;
	my ($batch_id) = @_;
	my $description = $self->{JobDescription};

	$self->log("cancel job $batch_id");

	$self->cd_fork_and_exec_cmd(undef,0,0,$qdel,$batch_id);

	if($? == 0 || $? == 153)
	{
		return { JOB_STATE => Globus::GRAM::JobState::FAILED }
	}

	return Globus::GRAM::Error::JOB_CANCEL_FAILED();
}

sub submit_to_batch_system
{
	my $self = shift;
	my ($submit_arg) = @_;
	my ($cpu_time,$wall_time) = split('\|',$submit_arg);
	my $description = $self->{JobDescription};
	# [prey]
	# Added internal_id variable
	my $internal_id = $description->jobid();
	my $tag = $description->cache_tag() || $ENV{GLOBUS_GRAM_JOB_CONTACT};
	my $cache_pgm = "$Globus::Core::Paths::bindir/globus-gass-cache";

	# [jfontan]
	# Real and fake stdout/stderr
	my $real_stdout;
	my $real_stderr;
	my $fake_stdout;
	my $fake_stderr;

	my $script_url = "$tag/sge_job_script"; 
	$self->cd_fork_and_exec_cmd(undef,0,0,$cache_pgm,"-add","-t",$tag,"-n",$script_url,"file:/dev/null");
	my $sge_job_script_name = `$cache_pgm -query -t $tag $script_url`;
	chomp($sge_job_script_name);
	if($sge_job_script_name eq "")
	{
		# [prey]
		# Call to dgas_record function		
		dgas_record("!", "$internal_id", "FAILED");
		return "FAILED";
	}

	my $sge_job_script = new IO::File($sge_job_script_name, '>');

	# [jfontan]
	# Changed EOF style by "" and added h_fsize (required by cesga queues)
	$sge_job_script->print("
#! /bin/sh
# SGE batch job script built by Globus job manager
#
#\$ -V
#\$ -cwd
#\$ -S /bin/sh
");

	# [prey]
	# Adding the extra parameters if it has been set in the Config file to a non-null value.
	if(defined $EXTRA && $EXTRA) { 
		$sge_job_script->print("#\$ -l ".$EXTRA."\n");
	}

	if(defined $description->email_address() && $description->email_address() ne '')
	{
		# [jfontan]
		# Mail wont be used with SGE
		#$sge_job_script->print("#PBS -M " . $description->email_address() . "\n");
	}
	if(defined $description->emailonabort() && $description->emailonabort() eq 'yes')
	{
		$email_when .= 'a';
	}
	if(defined $description->emailonexecution() && $description->emailonexecution() eq 'yes')
	{
		$email_when .= 'b';
	}
	if(defined $description->emailontermination() && $description->emailontermination() eq 'yes')
	{
		$email_when .= 'e';
	}
	if($email_when eq '')
	{
		$email_when = 'n';
	}
    
	# [jfontan]
	# Mail wont be used with SGE
	#$sge_job_script->print("#PBS -m $email_when\n");

	if(defined $description->queue() && $description->queue() ne '')
	{
		$sge_job_script->print("#\$ -q ". $description->queue() . "\n");
	}
	if(defined $description->project() && $description->project() ne '')
	{
		# [jfontan]
		# Project wont be used with SGE
		#$sge_job_script->print("#PBS -A " . $description->project() . "\n");
	}

	if($cpu_time != 0)
	{
		my $total_cpu_time;

		if($description->jobtype() eq 'multiple')
		{
			$total_cpu_time = $cpu_time * $description->count();
		}
		else
		{
			$total_cpu_time = $cpu_time;
		}
		#$sge_job_script->print("#PBS -l pcput=${cpu_time}:00\n");
		$sge_job_script->print("#\$ -l s_rt=${total_cpu_time}:00\n");
	} else {
		# [prey]
		# Adding s_rt if it has been set in the Config file to a non-null value.
		if(defined $S_RT && $S_RT) {
			$sge_job_script->print("#\$ -l s_rt=".$S_RT."\n");
		}
		#$sge_job_script->print("#\$ -l s_rt=10:00:00\n");
	}

	if($wall_time != 0)
	{
		# [jfontan]
		# Walltime wont be used with SGE
		#$sge_job_script->print("#PBS -l walltime=${wall_time}:00\n");
	}

	if(defined $description->max_memory() && $description->max_memory() != 0)
	{
		my $max_memory;

		if($description->jobtype() eq 'multiple')
		{
			$max_memory = $description->max_memory() * $description->count;
		}
		else
		{
			$max_memory = $description->max_memory();
		}
		$sge_job_script->print("#\$ -l s_vmem=${max_memory}M\n");
	} else {
                # [prey]
                # Adding s_vmem if it has been set in the Config file to a non-null value.
                if(defined $S_VMEM && $S_VMEM) {
                        $sge_job_script->print("#\$ -l s_vmem=".$S_VMEM."\n");
                }
		#$sge_job_script->print("#\$ -l s_vmem=512M\n");
	}

	chomp(my $my_hostname = `hostname -f`);
	mkdir '.lcgjm', 0700;
	chomp(my $pwd=`pwd`);
	chomp(my $cache_export_dir = `mktemp -d $pwd/.lcgjm/globus-cache-export.XXXXXX`);

	$self->helper_init_cache_export_url($cache_export_dir);

	# [jfontan]
	# Deal with stdout/stderr. SGE can not run jobs if it can not write to
	# stdout/stderr before starting the job, and $cache_export_dir does not
	# exist in worker nodes
	# my $r_number=int(rand(65536));
	$real_stdout=$cache_export_dir."/batch.out";
	$real_stderr=$cache_export_dir."/batch.err";
	# [ASG]
	# A best option for stdout/stderr names is to use the proper SGE names.
	#$fake_stdout=$pwd."/.out.".$r_number;
	#$fake_stderr=$pwd."/.err.".$r_number;

	$fake_stdout='$SGE_STDOUT_PATH';
	$fake_stderr='$SGE_STDERR_PATH';


# OLD LCGPBS LINES
#    $sge_job_script->print("#\$ -o " . $cache_export_dir."/batch.out" . "\n");
#    $sge_job_script->print("#\$ -e " . $cache_export_dir."/batch.err" . "\n");

#[ASG]
# Using stdout/stderr SGE by default.
#	$sge_job_script->print("#\$ -o " . $fake_stdout . "\n");
#	$sge_job_script->print("#\$ -e " . $fake_stderr . "\n");


	$sge_job_script->print("#\$ -r n\n");

	my @tmp_list = split("/",$cache_export_dir);
	my $gpg_file = pop(@tmp_list);
	$gpg_file .= ".gpg";

	# [jfontan]
	# stagein is not implemented in SGE, have to deal with that other way (X509SELF)
	#$sge_job_script->print("#PBS -W stagein=".$gpg_file."@".$my_hostname.":".$cache_export_dir."/".$gpg_file."\n");

	if(defined $description->host_count() && $description->host_count() != 0)
	{
		$sge_job_script->print("#\$ -l num_proc=" .
					$description->host_count().
					"\n");
	}
	elsif($cluster && $cpu_per_node != 0)
	{
		$sge_job_script->print("#\$ -l num_proc=" .
					POSIX::ceil($description->count /
					$cpu_per_node).
					"\n");
	}
	else
	{
                # [prey]
                # Adding num_proc if it has been set in the Config file to a non-null value.
                if(defined $NUM_PROC && $NUM_PROC) {
                        $sge_job_script->print("#\$ -l num_proc=".$NUM_PROC."\n");
                }
		#$sge_job_script->print("#\$ -l num_proc=1\n");
	}

	my @library_vars=('LD_LIBRARY_PATH');
	if($Config::Config{osname} eq 'irix')
	{
		push(@library_vars,'LD_LIBRARYN32_PATH');
		push(@library_vars,'LD_LIBRARY64_PATH');
	}

	my $rsh_env = "";
	my $local_x509 = '-';
	my @new_env;

	foreach my $tuple ($description->environment())
	{
		$tuple->[0] =~ s/"/\\\"/g;
		$tuple->[1] =~ s/"/\\\"/g;
		$self->helper_armour(\$tuple->[0]);
		$self->helper_armour(\$tuple->[1]);

		push(@new_env, $tuple->[0] . "=" . '"' . $tuple->[1] . '"');
		$local_x509 = $tuple->[1] if $tuple->[0] eq 'X509_USER_PROXY';

		$rsh_env .= $tuple->[0] . "=\"" . $tuple->[1] . "\"\n"
						.  "export " . $tuple->[0] . "\n";
	}

	# [jfontan]
	# Add environment directly in the script.
	# Changed the method to exporting the environment variables directly in the
	# script as SGE had some troubles using -v parameter with too much variables

	#$sge_job_script->print("#\$ -v " .  join(',', @new_env) . "\n");

	foreach my $env_line (@new_env)
	{
		$sge_job_script->print("export " . $env_line . "\n");
	}


	if (defined $description->library_path() && $description->library_path() ne '')
	{
		my @library_path;
		if (ref $description->library_path())
		{
			foreach my $tuple ($description->library_path())
			{
				push(@library_path,@$tuple);
			}
		}
		else
		{
			@library_path = ($description->library_path());
		}
		my $library_path = join(":",@library_path);
		foreach my $lib_var (@library_vars)
		{
			# [jfontan]
			# Changed EOF by ""
			$sge_job_script->print("
            if test \"X\${$lib_var}\" != \"X\"; then
                $lib_var=\"\${$lib_var}:$library_path\"
            else
                $lib_var=\"$library_path\"
            fi
            export $lib_var
");
			$rsh_env .= "if test \"X\${$lib_var}\" != \"X\"; then\n";
			$rsh_env .= "$lib_var=\"\${$lib_var}:$library_path\"\n";
			$rsh_env .= "else\n";
			$rsh_env .= "$lib_var=\"$library_path\"\n";
			$rsh_env .= "fi\n";
			$rsh_env .= "export $lib_var\n";
		}
	}

	# [jfontan]
	# Here we can have a problem if the first parameter is "0"
	# [TODO] fix it
	my $args="";
	my @arguments = $description->arguments();
	if($arguments[0])
	{
		foreach my $arg (@arguments)
		{
			$self->log("Transforming argument \"$arg\"\n");
			$self->helper_armour(\$arg);
			$self->log("Transformed to \"$arg\"\n");

			$args .= '"' . $arg . '" ';
		}
	}

#    if($description->jobtype() eq "mpi")
#    {
#        $sge_job_script->print("\n#Change to directory requested by user\n");
#        $sge_job_script->print('cd ' . $description->directory() . "\n");
#
#	$sge_job_script->print("$mpirun -np " . $description->count() . " ");
#
#	if($cluster)
#	{
#	    $sge_job_script->print(" -machinefile \$PBS_NODEFILE ");
#	}
#	$sge_job_script->print($description->executable()
#	                       . " $args < "
#			       . $description->stdin() . "\n");
#    }
#    elsif...

	# [jfontan]
	# Link fake stdout/stderr to the real one so the script can send back the results
	$sge_job_script->print("mkdir -p $cache_export_dir\n");
	#[ASG] $fake_stdout and $fake_stderr already created by SGE
	#$sge_job_script->print("touch $fake_stdout\n");
	$sge_job_script->print("ln -s $fake_stdout $real_stdout\n");
	#$sge_job_script->print("touch $fake_stderr\n");
	$sge_job_script->print("ln -s $fake_stderr $real_stderr\n");
		
	# [jfontan]
	# Add grid environment. This is needed in our cluster because WN are not
	# installed the standard way. This is not needed for standard LCG WN.
#	$sge_job_script->print(". /opt/cesga/lcg/etc/profile.d/grid_env.sh\n");
	# [prey]
	# Adding the grid enviroment if it has been set in the Config file to a non-null value.
	if(defined $GRID_ENV && $GRID_ENV) {
		$sge_job_script->print(". ".$GRID_ENV."\n");
	}

	@tmp_list=();
	$self->helper_get_from_tmp_file("scratch",\@tmp_list);
	my $gram_scratch_dir = (scalar(@tmp_list)>0) ? $tmp_list[0] : '-';

	# [jfontan]
	# Changed method to send the proxy from GPG to SELF (embeded in the script)

	if(($description->jobtype() eq 'multiple' && !$cluster) || $description->jobtype() eq 'single')
	{
#        $self->helper_write_non_mpi_script($sge_job_script,$my_hostname,$local_x509,
#                                           "X509GPG",$cache_export_dir,$gram_scratch_dir,"NONE",$args);
		$self->helper_write_non_mpi_script($sge_job_script,$my_hostname,$local_x509,
						"X509SELF",$cache_export_dir,$gram_scratch_dir,"NONE",$args);
	}
	elsif($description->jobtype() eq 'multiple')
	{
#        $self->helper_write_fake_mpi_script($sge_job_script,$rsh_env,$my_hostname,$local_x509,
#                                           "X509GPG",$cache_export_dir,$gram_scratch_dir,"NONE",$args);
		$self->helper_write_fake_mpi_script($sge_job_script,$rsh_env,$my_hostname,$local_x509,
						"X509SELF",$cache_export_dir,$gram_scratch_dir,"NONE",$args);

		my $count = $description->count;
				
		# [jfontan]
		# Changed EOF by ""
		# [TODO] This wont work but multiple jobs are not a priority
		$sge_job_script->print("
	hosts=\`cat \$PBS_NODEFILE\`;
	counter=0
	while test \$counter -lt $count; do
	    for host in \$hosts; do
		if test \$counter -lt $count; then
                    rsh_bootstrap=\`$remote_shell \$host mktemp /tmp/rsh_bootstrap.XXXXXX\`
                    $remote_shell \$host \"cat > \$rsh_bootstrap; chmod 700 \$rsh_bootstrap;\\
                                          \$rsh_bootstrap; rm \$rsh_bootstrap\" < \$local_bootstrap &
		    counter=\`expr \$counter + 1\`
		else
		    break
		fi
	    done
	done
	wait
        rm \$local_bootstrap
");
	}

	# [jfontan]
	# Delete cache dir and stdout/stderr
	# After the job completes stdout/stderr files are not deleted so $cache_export_dir
	# will remain in the home diretory.
	$sge_job_script->print("rm -rf ".$cache_export_dir."\n");
		
	# [jfontan]	
	# Could not delete this files until the job finishes :(
	# ...this is a job for... epilog!
	# This files are referenced by SGE_STDOUT_PATH and SGE_STDERR_PATH
	#$sge_job_script->print("rm -f ".$fake_stdout."\n");
	#$sge_job_script->print("rm -f ".$fake_stderr."\n");

	$sge_job_script->close();

	my $errfile;
	if($description->logfile() ne "")
	{
		$errfile = "2>>" . $description->logfile();
	}

	$self->helper_cache_export();

	$self->helper_remove_tmp_file();

	# [prey]
	# Call to dgas_record function
	if (! -e $local_x509) {
		dgas_record("!", "$internal_id", "FAILED");
		return "FAILED"
	}

	symlink $local_x509,$cache_export_dir."/.emergency-x509";
	symlink $sge_job_script_name,$cache_export_dir."/.submit-script";
	$self->cd_fork_and_exec_cmd(undef,0,0,"/bin/cp",$local_x509,$cache_export_dir."/.emergency-x509.ref");
  
	my $batch_id;
	my $submit_status;
	my $attempts=0;
	do
	{
		#chomp($batch_id = `$qsub < $sge_job_script_name 2>&1 > $errfile`);

		# [jfontan]
		# DEBUG: copy script to /tmp
		#system("cp -f $sge_job_script_name /tmp/sge_job_script");
		chomp($tmp_batch_id = `export SGE_ROOT=$sge_base_path ;$qsub < $sge_job_script_name`);
		$submit_status = $?;
		# [jfontan]
		# The way SGE qsub tells the job number is different to PBS
		if($tmp_batch_id =~ /^Your job (\d*)/) {
			$batch_id = $1;
		}
				
		$self->log("Batch ID: " . $batch_id);
				
		if ($submit_status != 0)
		{
			$batch_id = undef;
			sleep 5;
		}
		$attempts++;
	} while($submit_status != 0 && $attempts<6);

	if (!defined $batch_id)
	{
		$batch_id = "FAILED";
	}

	# [prey]
	# Call to dgas_record function
	dgas_record("!", "$internal_id", "$batch_id");
	return $batch_id;
}


1;
